@@ -33,6 +33,8 @@
 
 KJ_BEGIN_HEADER
 
+#include "list.h"
+
 namespace kj {
 namespace _ {  // private
 
@@ -106,6 +108,32 @@ inline void convertToReturn(ExceptionOr<Void>&& result) {
   }
 }
 
+class TraceBuilder {
+  // Helper for methods that build a call trace.
+public:
+  TraceBuilder(ArrayPtr<void*> space)
+      : start(space.begin()), current(space.begin()), limit(space.end()) {}
+
+  inline void add(void* addr) {
+    if (current < limit) {
+      *current++ = addr;
+    }
+  }
+
+  inline bool full() const { return current == limit; }
+
+  ArrayPtr<void*> finish() {
+    return arrayPtr(start, current);
+  }
+
+  String toString();
+
+private:
+  void** start;
+  void** current;
+  void** limit;
+};
+
 class Event {
   // An event waiting to be executed.  Not for direct use by applications -- promises use this
   // internally.
@@ -142,12 +170,17 @@ public:
   // If the event is armed but hasn't fired, cancel it. (Destroying the event does this
   // implicitly.)
 
-  kj::String trace();
-  // Dump debug info about this event.
+  virtual void traceEvent(TraceBuilder& builder) = 0;
+  // Build a trace of the callers leading up to this event. `builder` will be populated with
+  // "return addresses" of the promise chain waiting on this event. The return addresses may
+  // actually the addresses of lambdas passed to .then(), but in any case, feeding them into
+  // addr2line should produce useful source code locations.
+  //
+  // `traceEvent()` may be called from an async signal handler while `fire()` is executing. It
+  // must not allocate nor take locks.
 
-  virtual _::PromiseNode* getInnerForTrace();
-  // If this event wraps a PromiseNode, get that node.  Used for debug tracing.
-  // Default implementation returns nullptr.
+  String traceEvent();
+  // Helper that builds a trace and stringifies it.
 
 protected:
   virtual Maybe<Own<Event>> fire() = 0;
@@ -190,9 +223,27 @@ public:
   // Can only be called once, and only after the node is ready.  Must be called directly from the
   // event loop, with no application code on the stack.
 
-  virtual PromiseNode* getInnerForTrace();
-  // If this node wraps some other PromiseNode, get the wrapped node.  Used for debug tracing.
-  // Default implementation returns nullptr.
+  virtual void tracePromise(TraceBuilder& builder, bool stopAtNextEvent) = 0;
+  // Build a trace of this promise chain, showing what it is currently waiting on.
+  //
+  // Since traces are ordered callee-before-caller, PromiseNode::tracePromise() should typically
+  // recurse to its child first, then after the child returns, add itself to the trace.
+  //
+  // If `stopAtNextEvent` is true, then the trace should stop as soon as it hits a PromiseNode that
+  // also implements Event, and should not trace that node or its children. This is used in
+  // conjuction with Event::traceEvent(). The chain of Events is often more sparse than the chain
+  // of PromiseNodes, because a TransformPromiseNode (which implements .then()) is not itself an
+  // Event. TransformPromiseNode instead tells its child node to directly notify its *parent* node
+  // when it is ready, and then TransformPromiseNode applies the .then() transformation during the
+  // call to .get().
+  //
+  // So, when we trace the chain of Events backwards, we end up hoping over segments of
+  // TransformPromiseNodes (and other similar types). In order to get those added to the trace,
+  // each Event must call back down the PromiseNode chain in the opposite direction, using this
+  // method.
+  //
+  // `tracePromise()` may be called from an async signal handler while `get()` is executing. It
+  // must not allocate nor take locks.
 
   template <typename T>
   static Own<PromiseNode> from(T&& promise) {
@@ -222,6 +273,10 @@ protected:
     // Arms the event if init() has already been called and makes future calls to init()
     // automatically arm the event.
 
+    inline void traceEvent(TraceBuilder& builder) {
+      if (event != nullptr && !builder.full()) event->traceEvent(builder);
+    }
+
   private:
     Event* event = nullptr;
   };
@@ -242,6 +297,7 @@ public:
   ~ImmediatePromiseNodeBase() noexcept(false);
 
   void onReady(Event* event) noexcept override;
+  void tracePromise(TraceBuilder& builder, bool stopAtNextEvent) override;
 };
 
 template <typename T>
@@ -277,7 +333,7 @@ public:
 
   void onReady(Event* event) noexcept override;
   void get(ExceptionOrValue& output) noexcept override;
-  PromiseNode* getInnerForTrace() override;
+  void tracePromise(TraceBuilder& builder, bool stopAtNextEvent) override;
 
 private:
   Own<PromiseNode> dependency;
@@ -312,17 +368,40 @@ private:
 
 #if __GNUC__ >= 8 && !__clang__
 // GCC 8's class-memaccess warning rightly does not like the memcpy()'s below, but there's no
-// "legal" way for us to extract the contetn of a PTMF so too bad.
+// "legal" way for us to extract the content of a PTMF so too bad.
 #pragma GCC diagnostic push
 #pragma GCC diagnostic ignored "-Wclass-memaccess"
+#if __GNUC__ >= 11
+// GCC 11's array-bounds is  similarly upset with us for digging into "private" implementation
+// details. But the format is well-defined by the ABI which cannot change so please just let us
+// do it kthx.
+#pragma GCC diagnostic ignored "-Warray-bounds"
+#endif
 #endif
 
+template <typename T, typename ReturnType, typename... ParamTypes>
+void* getMethodStartAddress(T& obj, ReturnType (T::*method)(ParamTypes...));
+template <typename T, typename ReturnType, typename... ParamTypes>
+void* getMethodStartAddress(const T& obj, ReturnType (T::*method)(ParamTypes...) const);
+// Given an object and a pointer-to-method, return the start address of the method's code. The
+// intent is that this address can be used in a trace; addr2line should map it to the start of
+// the function's definition. For virtual methods, this does a vtable lookup on `obj` to determine
+// the address of the specific implementation (otherwise, `obj` wouldn't be needed).
+//
+// Note that if the method is overloaded or is a template, you will need to explicitly specify
+// the param and return types, otherwise the compiler won't know which overload / template
+// specialization you are requesting.
+
 class PtmfHelper {
-  // This class is a private helper for GetFunctorStartAddress. The class represents the internal
-  // representation of a pointer-to-member-function.
+  // This class is a private helper for GetFunctorStartAddress and getMethodStartAddress(). The
+  // class represents the internal representation of a pointer-to-member-function.
 
   template <typename... ParamTypes>
   friend struct GetFunctorStartAddress;
+  template <typename T, typename ReturnType, typename... ParamTypes>
+  friend void* getMethodStartAddress(T& obj, ReturnType (T::*method)(ParamTypes...));
+  template <typename T, typename ReturnType, typename... ParamTypes>
+  friend void* getMethodStartAddress(const T& obj, ReturnType (T::*method)(ParamTypes...) const);
 
 #if __GNUG__
 
@@ -330,7 +409,7 @@ class PtmfHelper {
   ptrdiff_t adj;
   // Layout of a pointer-to-member-function used by GCC and compatible compilers.
 
-  void* apply(void* obj) {
+  void* apply(const void* obj) {
 #if defined(__arm__) || defined(__mips__) || defined(__aarch64__)
     if (adj & 1) {
       ptrdiff_t voff = (ptrdiff_t)ptr;
@@ -353,7 +432,7 @@ class PtmfHelper {
 
 #else  // __GNUG__
 
-  void* apply(void* obj) { return nullptr; }
+  void* apply(const void* obj) { return nullptr; }
   // TODO(port):  PTMF instruction address extraction
 
 #define BODY return PtmfHelper{}
@@ -382,6 +461,15 @@ class PtmfHelper {
 #pragma GCC diagnostic pop
 #endif
 
+template <typename T, typename ReturnType, typename... ParamTypes>
+void* getMethodStartAddress(T& obj, ReturnType (T::*method)(ParamTypes...)) {
+  return PtmfHelper::from<ReturnType, T, ParamTypes...>(method).apply(&obj);
+}
+template <typename T, typename ReturnType, typename... ParamTypes>
+void* getMethodStartAddress(const T& obj, ReturnType (T::*method)(ParamTypes...) const) {
+  return PtmfHelper::from<ReturnType, T, ParamTypes...>(method).apply(&obj);
+}
+
 template <typename... ParamTypes>
 struct GetFunctorStartAddress {
   // Given a functor (any object defining operator()), return the start address of the function,
@@ -414,7 +502,7 @@ public:
 
   void onReady(Event* event) noexcept override;
   void get(ExceptionOrValue& output) noexcept override;
-  PromiseNode* getInnerForTrace() override;
+  void tracePromise(TraceBuilder& builder, bool stopAtNextEvent) override;
 
 private:
   Own<PromiseNode> dependency;
@@ -435,9 +523,9 @@ class TransformPromiseNode final: public TransformPromiseNodeBase {
   // function (implements `then()`).
 
 public:
-  TransformPromiseNode(Own<PromiseNode>&& dependency, Func&& func, ErrorFunc&& errorHandler)
-      : TransformPromiseNodeBase(kj::mv(dependency),
-            GetFunctorStartAddress<DepT&&>::apply(func)),
+  TransformPromiseNode(Own<PromiseNode>&& dependency, Func&& func, ErrorFunc&& errorHandler,
+                       void* continuationTracePtr)
+      : TransformPromiseNodeBase(kj::mv(dependency), continuationTracePtr),
         func(kj::fwd<Func>(func)), errorHandler(kj::fwd<ErrorFunc>(errorHandler)) {}
 
   ~TransformPromiseNode() noexcept(false) {
@@ -485,7 +573,7 @@ public:
 
   // implements PromiseNode ------------------------------------------
   void onReady(Event* event) noexcept override;
-  PromiseNode* getInnerForTrace() override;
+  void tracePromise(TraceBuilder& builder, bool stopAtNextEvent) override;
 
 protected:
   inline ExceptionOrValue& getHubResultRef();
@@ -505,6 +593,11 @@ private:
 
 template <typename T> T copyOrAddRef(T& t) { return t; }
 template <typename T> Own<T> copyOrAddRef(Own<T>& t) { return t->addRef(); }
+template <typename T> Maybe<Own<T>> copyOrAddRef(Maybe<Own<T>>& t) {
+  return t.map([](Own<T>& ptr) {
+    return ptr->addRef();
+  });
+}
 
 template <typename T>
 class ForkBranch final: public ForkBranchBase {
@@ -565,7 +658,7 @@ private:
   // Tail becomes null once the inner promise is ready and all branches have been notified.
 
   Maybe<Own<Event>> fire() override;
-  _::PromiseNode* getInnerForTrace() override;
+  void traceEvent(TraceBuilder& builder) override;
 
   friend class ForkBranchBase;
 };
@@ -622,7 +715,7 @@ public:
   void onReady(Event* event) noexcept override;
   void setSelfPointer(Own<PromiseNode>* selfPtr) noexcept override;
   void get(ExceptionOrValue& output) noexcept override;
-  PromiseNode* getInnerForTrace() override;
+  void tracePromise(TraceBuilder& builder, bool stopAtNextEvent) override;
 
 private:
   enum State {
@@ -640,6 +733,7 @@ private:
   Own<PromiseNode>* selfPtr = nullptr;
 
   Maybe<Own<Event>> fire() override;
+  void traceEvent(TraceBuilder& builder) override;
 };
 
 template <typename T>
@@ -671,7 +765,7 @@ public:
 
   void onReady(Event* event) noexcept override;
   void get(ExceptionOrValue& output) noexcept override;
-  PromiseNode* getInnerForTrace() override;
+  void tracePromise(TraceBuilder& builder, bool stopAtNextEvent) override;
 
 private:
   class Branch: public Event {
@@ -683,11 +777,13 @@ private:
     // Returns true if this is the side that finished.
 
     Maybe<Own<Event>> fire() override;
-    _::PromiseNode* getInnerForTrace() override;
+    void traceEvent(TraceBuilder& builder) override;
 
   private:
     ExclusiveJoinPromiseNode& joinNode;
     Own<PromiseNode> dependency;
+
+    friend class ExclusiveJoinPromiseNode;
   };
 
   Branch left;
@@ -705,7 +801,7 @@ public:
 
   void onReady(Event* event) noexcept override final;
   void get(ExceptionOrValue& output) noexcept override final;
-  PromiseNode* getInnerForTrace() override final;
+  void tracePromise(TraceBuilder& builder, bool stopAtNextEvent) override final;
 
 protected:
   virtual void getNoError(ExceptionOrValue& output) noexcept = 0;
@@ -722,7 +818,7 @@ private:
     ~Branch() noexcept(false);
 
     Maybe<Own<Event>> fire() override;
-    _::PromiseNode* getInnerForTrace() override;
+    void traceEvent(TraceBuilder& builder) override;
 
     Maybe<Exception> getPart();
     // Calls dependency->get(output).  If there was an exception, return it.
@@ -731,6 +827,8 @@ private:
     ArrayJoinPromiseNodeBase& joinNode;
     Own<PromiseNode> dependency;
     ExceptionOrValue& output;
+
+    friend class ArrayJoinPromiseNodeBase;
   };
 
   Array<Branch> branches;
@@ -783,7 +881,7 @@ public:
   EagerPromiseNodeBase(Own<PromiseNode>&& dependency, ExceptionOrValue& resultRef);
 
   void onReady(Event* event) noexcept override;
-  PromiseNode* getInnerForTrace() override;
+  void tracePromise(TraceBuilder& builder, bool stopAtNextEvent) override;
 
 private:
   Own<PromiseNode> dependency;
@@ -792,6 +890,7 @@ private:
   ExceptionOrValue& resultRef;
 
   Maybe<Own<Event>> fire() override;
+  void traceEvent(TraceBuilder& builder) override;
 };
 
 template <typename T>
@@ -820,6 +919,7 @@ Own<PromiseNode> spark(Own<PromiseNode>&& node) {
 class AdapterPromiseNodeBase: public PromiseNode {
 public:
   void onReady(Event* event) noexcept override;
+  void tracePromise(TraceBuilder& builder, bool stopAtNextEvent) override;
 
 protected:
   inline void setReady() {
@@ -877,7 +977,8 @@ class FiberBase: public PromiseNode, private Event {
   // Base class for the outer PromiseNode representing a fiber.
 
 public:
-  FiberBase(size_t stackSize, _::ExceptionOrValue& result);
+  explicit FiberBase(size_t stackSize, _::ExceptionOrValue& result);
+  explicit FiberBase(const FiberPool& pool, _::ExceptionOrValue& result);
   ~FiberBase() noexcept(false);
 
   void start() { armDepthFirst(); }
@@ -886,7 +987,7 @@ public:
   class WaitDoneEvent;
 
   void onReady(_::Event* event) noexcept override;
-  PromiseNode* getInnerForTrace() override;
+  void tracePromise(TraceBuilder& builder, bool stopAtNextEvent) override;
 
 protected:
   bool isFinished() { return state == FINISHED; }
@@ -895,31 +996,19 @@ protected:
 private:
   enum { WAITING, RUNNING, CANCELED, FINISHED } state;
 
-  size_t stackSize;
-
-#if _WIN32 || __CYGWIN__
-  void* osFiber;
-#elif !__BIONIC__
-  struct Impl;
-  Impl& impl;
-#endif
-
   _::PromiseNode* currentInner = nullptr;
   OnReadyEvent onReadyEvent;
+  Own<FiberStack> stack;
   _::ExceptionOrValue& result;
 
   void run();
   virtual void runImpl(WaitScope& waitScope) = 0;
 
-  struct StartRoutine;
-
-  void switchToFiber();
-  void switchToMain();
-
   Maybe<Own<Event>> fire() override;
+  void traceEvent(TraceBuilder& builder) override;
   // Implements Event. Each time the event is fired, switchToFiber() is called.
 
-  friend class WaitScope;
+  friend class FiberStack;
   friend void _::waitImpl(Own<_::PromiseNode>&& node, _::ExceptionOrValue& result,
                           WaitScope& waitScope);
   friend bool _::pollImpl(_::PromiseNode& node, WaitScope& waitScope);
@@ -928,7 +1017,8 @@ private:
 template <typename Func>
 class Fiber final: public FiberBase {
 public:
-  Fiber(size_t stackSize, Func&& func): FiberBase(stackSize, result), func(kj::fwd<Func>(func)) {}
+  explicit Fiber(size_t stackSize, Func&& func): FiberBase(stackSize, result), func(kj::fwd<Func>(func)) {}
+  explicit Fiber(const FiberPool& pool, Func&& func): FiberBase(pool, result), func(kj::fwd<Func>(func)) {}
   ~Fiber() noexcept(false) { destroy(); }
 
   typedef FixVoid<decltype(kj::instance<Func&>()(kj::instance<WaitScope&>()))> ResultType;
@@ -965,9 +1055,11 @@ template <typename Func, typename ErrorFunc>
 PromiseForResult<Func, T> Promise<T>::then(Func&& func, ErrorFunc&& errorHandler) {
   typedef _::FixVoid<_::ReturnType<Func, T>> ResultT;
 
+  void* continuationTracePtr = _::GetFunctorStartAddress<_::FixVoid<T>&&>::apply(func);
   Own<_::PromiseNode> intermediate =
       heap<_::TransformPromiseNode<ResultT, _::FixVoid<T>, Func, ErrorFunc>>(
-          kj::mv(node), kj::fwd<Func>(func), kj::fwd<ErrorFunc>(errorHandler));
+          kj::mv(node), kj::fwd<Func>(func), kj::fwd<ErrorFunc>(errorHandler),
+          continuationTracePtr);
   auto result = _::PromiseNode::to<_::ChainPromises<_::ReturnType<Func, T>>>(
       _::maybeChain(kj::mv(intermediate), implicitCast<ResultT*>(nullptr)));
   return _::maybeReduce(kj::mv(result), false);
@@ -1008,8 +1100,18 @@ Promise<T> Promise<T>::catch_(ErrorFunc&& errorHandler) {
   // Func is being filled in automatically. We want to make sure ErrorFunc can return a Promise,
   // but we don't want the extra overhead of promise chaining if ErrorFunc doesn't actually
   // return a promise. So we make our Func return match ErrorFunc.
-  return then(_::IdentityFunc<decltype(errorHandler(instance<Exception&&>()))>(),
-              kj::fwd<ErrorFunc>(errorHandler));
+  typedef _::IdentityFunc<decltype(errorHandler(instance<Exception&&>()))> Func;
+  typedef _::FixVoid<_::ReturnType<Func, T>> ResultT;
+
+  // The reason catch_() isn't simply implemented in terms of then() is because we want the trace
+  // pointer to be based on ErrorFunc rather than Func.
+  void* continuationTracePtr = _::GetFunctorStartAddress<kj::Exception&&>::apply(errorHandler);
+  Own<_::PromiseNode> intermediate =
+      heap<_::TransformPromiseNode<ResultT, _::FixVoid<T>, Func, ErrorFunc>>(
+          kj::mv(node), Func(), kj::fwd<ErrorFunc>(errorHandler), continuationTracePtr);
+  auto result = _::PromiseNode::to<_::ChainPromises<_::ReturnType<Func, T>>>(
+      _::maybeChain(kj::mv(intermediate), implicitCast<ResultT*>(nullptr)));
+  return _::maybeReduce(kj::mv(result), false);
 }
 
 template <typename T>
@@ -1034,6 +1136,11 @@ Promise<T> ForkedPromise<T>::addBranch() {
   return hub->addBranch();
 }
 
+template <typename T>
+bool ForkedPromise<T>::hasBranches() {
+  return hub->isShared();
+}
+
 template <typename T>
 _::SplitTuplePromise<T> Promise<T>::split() {
   return refcounted<_::ForkHub<_::FixVoid<T>>>(kj::mv(node))->split();
@@ -1091,6 +1198,42 @@ inline PromiseForResult<Func, void> evalNow(Func&& func) {
   return result;
 }
 
+template <typename Func>
+struct RetryOnDisconnect_ {
+  static inline PromiseForResult<Func, void> apply(Func&& func) {
+    return evalLater([func = kj::mv(func)]() mutable -> PromiseForResult<Func, void> {
+      auto promise = evalNow(func);
+      return promise.catch_([func = kj::mv(func)](kj::Exception&& e) mutable -> PromiseForResult<Func, void> {
+        if (e.getType() == kj::Exception::Type::DISCONNECTED) {
+          return func();
+        } else {
+          return kj::mv(e);
+        }
+      });
+    });
+  }
+};
+template <typename Func>
+struct RetryOnDisconnect_<Func&> {
+  // Specialization for references. Needed because the syntax for capturing references in a
+  // lambda is different. :(
+  static inline PromiseForResult<Func, void> apply(Func& func) {
+    auto promise = evalLater(func);
+    return promise.catch_([&func](kj::Exception&& e) -> PromiseForResult<Func, void> {
+      if (e.getType() == kj::Exception::Type::DISCONNECTED) {
+        return func();
+      } else {
+        return kj::mv(e);
+      }
+    });
+  }
+};
+
+template <typename Func>
+inline PromiseForResult<Func, void> retryOnDisconnect(Func&& func) {
+  return RetryOnDisconnect_<Func>::apply(kj::fwd<Func>(func));
+}
+
 template <typename Func>
 inline PromiseForResult<Func, WaitScope&> startFiber(size_t stackSize, Func&& func) {
   typedef _::FixVoid<_::ReturnType<Func, WaitScope&>> ResultT;
@@ -1102,6 +1245,17 @@ inline PromiseForResult<Func, WaitScope&> startFiber(size_t stackSize, Func&& fu
   return _::maybeReduce(kj::mv(result), false);
 }
 
+template <typename Func>
+inline PromiseForResult<Func, WaitScope&> FiberPool::startFiber(Func&& func) const {
+  typedef _::FixVoid<_::ReturnType<Func, WaitScope&>> ResultT;
+
+  Own<_::FiberBase> intermediate = kj::heap<_::Fiber<Func>>(*this, kj::fwd<Func>(func));
+  intermediate->start();
+  auto result = _::PromiseNode::to<_::ChainPromises<_::ReturnType<Func, WaitScope&>>>(
+      _::maybeChain(kj::mv(intermediate), implicitCast<ResultT*>(nullptr)));
+  return _::maybeReduce(kj::mv(result), false);
+}
+
 template <typename T>
 template <typename ErrorFunc>
 void Promise<T>::detach(ErrorFunc&& errorHandler) {
@@ -1125,8 +1279,28 @@ Promise<Array<T>> joinPromises(Array<Promise<T>>&& promises) {
 
 namespace _ {  // private
 
+class WeakFulfillerBase: protected kj::Disposer {
+protected:
+  WeakFulfillerBase(): inner(nullptr) {}
+  virtual ~WeakFulfillerBase() noexcept(false) {}
+
+  template <typename T>
+  inline PromiseFulfiller<T>* getInner() {
+    return static_cast<PromiseFulfiller<T>*>(inner);
+  };
+  template <typename T>
+  inline void setInner(PromiseFulfiller<T>* ptr) {
+    inner = ptr;
+  };
+
+private:
+  mutable PromiseRejector* inner;
+
+  void disposeImpl(void* pointer) const override;
+};
+
 template <typename T>
-class WeakFulfiller final: public PromiseFulfiller<T>, private kj::Disposer {
+class WeakFulfiller final: public PromiseFulfiller<T>, public WeakFulfillerBase {
   // A wrapper around PromiseFulfiller which can be detached.
   //
   // There are a couple non-trivialities here:
@@ -1149,54 +1323,37 @@ public:
   }
 
   void fulfill(FixVoid<T>&& value) override {
-    if (inner != nullptr) {
-      inner->fulfill(kj::mv(value));
+    if (getInner<T>() != nullptr) {
+      getInner<T>()->fulfill(kj::mv(value));
     }
   }
 
   void reject(Exception&& exception) override {
-    if (inner != nullptr) {
-      inner->reject(kj::mv(exception));
+    if (getInner<T>() != nullptr) {
+      getInner<T>()->reject(kj::mv(exception));
     }
   }
 
   bool isWaiting() override {
-    return inner != nullptr && inner->isWaiting();
+    return getInner<T>() != nullptr && getInner<T>()->isWaiting();
   }
 
   void attach(PromiseFulfiller<T>& newInner) {
-    inner = &newInner;
+    setInner<T>(&newInner);
   }
 
   void detach(PromiseFulfiller<T>& from) {
-    if (inner == nullptr) {
+    if (getInner<T>() == nullptr) {
       // Already disposed.
       delete this;
     } else {
-      KJ_IREQUIRE(inner == &from);
-      inner = nullptr;
+      KJ_IREQUIRE(getInner<T>() == &from);
+      setInner<T>(nullptr);
     }
   }
 
 private:
-  mutable PromiseFulfiller<T>* inner;
-
-  WeakFulfiller(): inner(nullptr) {}
-
-  void disposeImpl(void* pointer) const override {
-    // TODO(perf): Factor some of this out so it isn't regenerated for every fulfiller type?
-
-    if (inner == nullptr) {
-      // Already detached.
-      delete this;
-    } else {
-      if (inner->isWaiting()) {
-        inner->reject(kj::Exception(kj::Exception::Type::FAILED, __FILE__, __LINE__,
-            kj::heapString("PromiseFulfiller was destroyed without fulfilling the promise.")));
-      }
-      inner = nullptr;
-    }
-  }
+  WeakFulfiller() {}
 };
 
 template <typename T>
@@ -1269,8 +1426,9 @@ namespace _ {  // (private)
 class XThreadEvent: private Event,         // it's an event in the target thread
                     public PromiseNode {   // it's a PromiseNode in the requesting thread
 public:
-  XThreadEvent(ExceptionOrValue& result, const Executor& targetExecutor)
-      : Event(targetExecutor.loop), result(result), targetExecutor(targetExecutor) {}
+  XThreadEvent(ExceptionOrValue& result, const Executor& targetExecutor, void* funcTracePtr);
+
+  void tracePromise(TraceBuilder& builder, bool stopAtNextEvent) override;
 
 protected:
   void ensureDoneOrCanceled();
@@ -1287,15 +1445,15 @@ protected:
 
 private:
   ExceptionOrValue& result;
+  void* funcTracePtr;
 
-  const Executor& targetExecutor;
+  kj::Own<const Executor> targetExecutor;
   Maybe<const Executor&> replyExecutor;  // If executeAsync() was used.
 
   kj::Maybe<Own<PromiseNode>> promiseNode;
   // Accessed only in target thread.
 
-  Maybe<XThreadEvent&> targetNext;
-  Maybe<XThreadEvent&>* targetPrev = nullptr;
+  ListLink<XThreadEvent> targetLink;
   // Membership in one of the linked lists in the target Executor's work list or cancel list. These
   // fields are protected by the target Executor's mutex.
 
@@ -1304,12 +1462,17 @@ private:
     // Object was never queued on another thread.
 
     QUEUED,
-    // Target thread has not yet dequeued the event from crossThreadRequests.events. The requesting
+    // Target thread has not yet dequeued the event from the state.start list. The requesting
     // thread can cancel execution by removing the event from the list.
 
     EXECUTING,
-    // Target thread has dequeued the event and is executing it. To cancel, the requesting thread
-    // must add the event to the crossThreadRequests.cancel list.
+    // Target thread has dequeued the event from state.start and moved it to state.executing. To
+    // cancel, the requesting thread must add the event to the state.cancel list and change the
+    // state to CANCELING.
+
+    CANCELING,
+    // Requesting thread is trying to cancel this event. The target thread will change the state to
+    // `DONE` once canceled.
 
     DONE
     // Target thread has completed handling this event and will not touch it again. The requesting
@@ -1319,8 +1482,7 @@ private:
   } state = UNUSED;
   // State, which is also protected by `targetExecutor`'s mutex.
 
-  Maybe<XThreadEvent&> replyNext;
-  Maybe<XThreadEvent&>* replyPrev = nullptr;
+  ListLink<XThreadEvent> replyLink;
   // Membership in `replyExecutor`'s reply list. Protected by `replyExecutor`'s mutex. The
   // executing thread places the event in the reply list near the end of the `EXECUTING` state.
   // Because the thread cannot lock two mutexes at once, it's possible that the reply executor
@@ -1333,6 +1495,23 @@ private:
   friend class kj::Executor;
 
   void done();
+  // Sets the state to `DONE` and notifies the originating thread that this event is done. Do NOT
+  // call under lock.
+
+  void sendReply();
+  // Notifies the originating thread that this event is done, but doesn't set the state to DONE
+  // yet. Do NOT call under lock.
+
+  void setDoneState();
+  // Assigns `state` to `DONE`, being careful to use an atomic-release-store if needed. This must
+  // only be called in the destination thread, and must either be called under lock, or the thread
+  // must take the lock and release it again shortly after setting the state (because some threads
+  // may be waiting on the DONE state using a conditional wait on the mutex). After calling
+  // setDoneState(), the destination thread MUST NOT touch this object ever again; it now belongs
+  // solely to the requesting thread.
+
+  void setDisconnected();
+  // Sets the result to a DISCONNECTED exception indicating that the target event loop exited.
 
   class DelayedDoneHack;
 
@@ -1340,6 +1519,8 @@ private:
   Maybe<Own<Event>> fire() override;
   // If called with promiseNode == nullptr, it's time to call execute(). If promiseNode != nullptr,
   // then it just indicated readiness and we need to get its result.
+
+  void traceEvent(TraceBuilder& builder) override;
 };
 
 template <typename Func, typename = _::FixVoid<_::ReturnType<Func, void>>>
@@ -1347,7 +1528,8 @@ class XThreadEventImpl final: public XThreadEvent {
   // Implementation for a function that does not return a Promise.
 public:
   XThreadEventImpl(Func&& func, const Executor& target)
-      : XThreadEvent(result, target), func(kj::fwd<Func>(func)) {}
+      : XThreadEvent(result, target, GetFunctorStartAddress<>::apply(func)),
+        func(kj::fwd<Func>(func)) {}
   ~XThreadEventImpl() noexcept(false) { ensureDoneOrCanceled(); }
 
   typedef _::FixVoid<_::ReturnType<Func, void>> ResultT;
@@ -1373,7 +1555,8 @@ class XThreadEventImpl<Func, Promise<T>> final: public XThreadEvent {
   // Implementation for a function that DOES return a Promise.
 public:
   XThreadEventImpl(Func&& func, const Executor& target)
-      : XThreadEvent(result, target), func(kj::fwd<Func>(func)) {}
+      : XThreadEvent(result, target, GetFunctorStartAddress<>::apply(func)),
+        func(kj::fwd<Func>(func)) {}
   ~XThreadEventImpl() noexcept(false) { ensureDoneOrCanceled(); }
 
   typedef _::FixVoid<_::UnwrapPromise<PromiseForResult<Func, void>>> ResultT;
@@ -1411,6 +1594,178 @@ PromiseForResult<Func, void> Executor::executeAsync(Func&& func) const {
   return _::PromiseNode::to<PromiseForResult<Func, void>>(kj::mv(event));
 }
 
+// -----------------------------------------------------------------------------
+
+namespace _ {  // (private)
+
+template <typename T>
+class XThreadFulfiller;
+
+class XThreadPaf: public PromiseNode {
+public:
+  XThreadPaf();
+  virtual ~XThreadPaf() noexcept(false);
+
+  class Disposer: public kj::Disposer {
+  public:
+    void disposeImpl(void* pointer) const override;
+  };
+  static const Disposer DISPOSER;
+
+  // implements PromiseNode ----------------------------------------------------
+  void onReady(Event* event) noexcept override;
+  void tracePromise(TraceBuilder& builder, bool stopAtNextEvent) override;
+
+private:
+  enum {
+    WAITING,
+    // Not yet fulfilled, and the waiter is still waiting.
+    //
+    // Starting from this state, the state may transition to either FULFILLING or CANCELED
+    // using an atomic compare-and-swap.
+
+    FULFILLING,
+    // The fulfiller thread atomically transitions the state from WAITING to FULFILLING when it
+    // wishes to fulfill the promise. By doing so, it guarantees that the `executor` will not
+    // disappear out from under it. It then fills in the result value, locks the executor mutex,
+    // adds the object to the executor's list of fulfilled XThreadPafs, changes the state to
+    // FULFILLED, and finally unlocks the mutex.
+    //
+    // If the waiting thread tries to cancel but discovers the object in this state, then it
+    // must perform a conditional wait on the executor mutex to await the state becoming FULFILLED.
+    // It can then delete the object.
+
+    FULFILLED,
+    // The fulfilling thread has completed filling in the result value and inserting the object
+    // into the waiting thread's executor event queue. Moreover, the fulfilling thread no longer
+    // holds any pointers to this object. The waiting thread is responsible for deleting it.
+
+    DISPATCHED,
+    // The object reached FULFILLED state, and then was dispatched from the waiting thread's
+    // executor's event queue. Therefore, the object is completely owned by the waiting thread with
+    // no need to lock anything.
+
+    CANCELED
+    // The waiting thread atomically transitions the state from WAITING to CANCELED if it is no
+    // longer listening. In this state, it is the fulfiller thread's responsibility to destroy the
+    // object.
+  } state;
+
+  const Executor& executor;
+  // Executor of the waiting thread. Only guaranteed to be valid when state is `WAITING` or
+  // `FULFILLING`. After any other state has been reached, this reference may be invalidated.
+
+  ListLink<XThreadPaf> link;
+  // In the FULFILLING/FULFILLED states, the object is placed in a linked list within the waiting
+  // thread's executor. In those states, these pointers are guarded by said executor's mutex.
+
+  OnReadyEvent onReadyEvent;
+
+  class FulfillScope;
+
+  static kj::Exception unfulfilledException();
+  // Construct appropriate exception to use to reject an unfulfilled XThreadPaf.
+
+  template <typename T>
+  friend class XThreadFulfiller;
+  friend Executor;
+};
+
+template <typename T>
+class XThreadPafImpl final: public XThreadPaf {
+public:
+  // implements PromiseNode ----------------------------------------------------
+  void get(ExceptionOrValue& output) noexcept override {
+    output.as<FixVoid<T>>() = kj::mv(result);
+  }
+
+private:
+  ExceptionOr<FixVoid<T>> result;
+
+  friend class XThreadFulfiller<T>;
+};
+
+class XThreadPaf::FulfillScope {
+  // Create on stack while setting `XThreadPafImpl<T>::result`.
+  //
+  // This ensures that:
+  // - Only one call is carried out, even if multiple threads try to fulfill concurrently.
+  // - The waiting thread is correctly signaled.
+public:
+  FulfillScope(XThreadPaf** pointer);
+  // Atomically nulls out *pointer and takes ownership of the pointer.
+
+  ~FulfillScope() noexcept(false);
+
+  KJ_DISALLOW_COPY(FulfillScope);
+
+  bool shouldFulfill() { return obj != nullptr; }
+
+  template <typename T>
+  XThreadPafImpl<T>* getTarget() { return static_cast<XThreadPafImpl<T>*>(obj); }
+
+private:
+  XThreadPaf* obj;
+};
+
+template <typename T>
+class XThreadFulfiller final: public CrossThreadPromiseFulfiller<T> {
+public:
+  XThreadFulfiller(XThreadPafImpl<T>* target): target(target) {}
+
+  ~XThreadFulfiller() noexcept(false) {
+    if (target != nullptr) {
+      reject(XThreadPaf::unfulfilledException());
+    }
+  }
+  void fulfill(FixVoid<T>&& value) const override {
+    XThreadPaf::FulfillScope scope(&target);
+    if (scope.shouldFulfill()) {
+      scope.getTarget<T>()->result = kj::mv(value);
+    }
+  }
+  void reject(Exception&& exception) const override {
+    XThreadPaf::FulfillScope scope(&target);
+    if (scope.shouldFulfill()) {
+      scope.getTarget<T>()->result.addException(kj::mv(exception));
+    }
+  }
+  bool isWaiting() const override {
+    KJ_IF_MAYBE(t, target) {
+#if _MSC_VER && !__clang__
+      // Just assume 1-byte loads are atomic... on what kind of absurd platform would they not be?
+      return t->state == XThreadPaf::WAITING;
+#else
+      return __atomic_load_n(&t->state, __ATOMIC_RELAXED) == XThreadPaf::WAITING;
+#endif
+    } else {
+      return false;
+    }
+  }
+
+private:
+  mutable XThreadPaf* target;  // accessed using atomic ops
+};
+
+template <typename T>
+class XThreadFulfiller<kj::Promise<T>> {
+public:
+  static_assert(sizeof(T) < 0,
+      "newCrosssThreadPromiseAndFulfiller<Promise<T>>() is not currently supported");
+  // TODO(someday): Is this worth supporting? Presumably, when someone calls `fulfill(somePromise)`,
+  //   then `somePromise` should be assumed to be a promise owned by the fulfilling thread, not
+  //   the waiting thread.
+};
+
+}  // namespace _ (private)
+
+template <typename T>
+PromiseCrossThreadFulfillerPair<T> newPromiseAndCrossThreadFulfiller() {
+  kj::Own<_::XThreadPafImpl<T>> node(new _::XThreadPafImpl<T>, _::XThreadPaf::DISPOSER);
+  auto fulfiller = kj::heap<_::XThreadFulfiller<T>>(node);
+  return { _::PromiseNode::to<_::ReducePromises<T>>(kj::mv(node)), kj::mv(fulfiller) };
+}
+
 }  // namespace kj
 
 KJ_END_HEADER
