@@ -42,25 +42,25 @@ public:
   inline explicit Vector(size_t capacity): builder(heapArrayBuilder<T>(capacity)) {}
   inline Vector(Array<T>&& array): builder(kj::mv(array)) {}
 
-  inline operator ArrayPtr<T>() { return builder; }
-  inline operator ArrayPtr<const T>() const { return builder; }
-  inline ArrayPtr<T> asPtr() { return builder.asPtr(); }
-  inline ArrayPtr<const T> asPtr() const { return builder.asPtr(); }
+  inline operator ArrayPtr<T>() KJ_LIFETIMEBOUND { return builder; }
+  inline operator ArrayPtr<const T>() const KJ_LIFETIMEBOUND { return builder; }
+  inline ArrayPtr<T> asPtr() KJ_LIFETIMEBOUND { return builder.asPtr(); }
+  inline ArrayPtr<const T> asPtr() const KJ_LIFETIMEBOUND { return builder.asPtr(); }
 
   inline size_t size() const { return builder.size(); }
   inline bool empty() const { return size() == 0; }
   inline size_t capacity() const { return builder.capacity(); }
-  inline T& operator[](size_t index) { return builder[index]; }
-  inline const T& operator[](size_t index) const { return builder[index]; }
+  inline T& operator[](size_t index) KJ_LIFETIMEBOUND { return builder[index]; }
+  inline const T& operator[](size_t index) const KJ_LIFETIMEBOUND { return builder[index]; }
 
-  inline const T* begin() const { return builder.begin(); }
-  inline const T* end() const { return builder.end(); }
-  inline const T& front() const { return builder.front(); }
-  inline const T& back() const { return builder.back(); }
-  inline T* begin() { return builder.begin(); }
-  inline T* end() { return builder.end(); }
-  inline T& front() { return builder.front(); }
-  inline T& back() { return builder.back(); }
+  inline const T* begin() const KJ_LIFETIMEBOUND { return builder.begin(); }
+  inline const T* end() const KJ_LIFETIMEBOUND { return builder.end(); }
+  inline const T& front() const KJ_LIFETIMEBOUND { return builder.front(); }
+  inline const T& back() const KJ_LIFETIMEBOUND { return builder.back(); }
+  inline T* begin() KJ_LIFETIMEBOUND { return builder.begin(); }
+  inline T* end() KJ_LIFETIMEBOUND { return builder.end(); }
+  inline T& front() KJ_LIFETIMEBOUND { return builder.front(); }
+  inline T& back() KJ_LIFETIMEBOUND { return builder.back(); }
 
   inline Array<T> releaseAsArray() {
     // TODO(perf):  Avoid a copy/move by allowing Array<T> to point to incomplete space?
@@ -75,15 +75,15 @@ public:
   template <typename U>
   inline bool operator!=(const U& other) const { return asPtr() != other; }
 
-  inline ArrayPtr<T> slice(size_t start, size_t end) {
+  inline ArrayPtr<T> slice(size_t start, size_t end) KJ_LIFETIMEBOUND {
     return asPtr().slice(start, end);
   }
-  inline ArrayPtr<const T> slice(size_t start, size_t end) const {
+  inline ArrayPtr<const T> slice(size_t start, size_t end) const KJ_LIFETIMEBOUND {
     return asPtr().slice(start, end);
   }
 
   template <typename... Params>
-  inline T& add(Params&&... params) {
+  inline T& add(Params&&... params) KJ_LIFETIMEBOUND {
     if (builder.isFull()) grow();
     return builder.add(kj::fwd<Params>(params)...);
   }
